"""  ServiceWall

Uses service definitions provided by jhansonxi
and implements them in a FireWall class, either to allow them for the local
subnetwork, or worldwide.

This firewall records the services you allow your computer to serve on a
specific network profile, switching between profiles as network provider
changes.

"""

import pickle
import json
import copy
import os
import subprocess
from collections import namedtuple

from servicewall import network_helpers
from servicewall import statefulfirewall

# from servicewall import service_helpers
# globals()["PortDef"] = service_helpers.PortDef
# globals()["ServiceDef"] = service_helpers.ServiceDef

from servicewall.service_helpers import PortDef, ServiceDef
globals()["PortDef"] = PortDef
globals()["ServiceDef"] = ServiceDef


def _systemctl(arg):
    try:
        retval = subprocess.check_output(['systemctl',
                                          arg,
                                          'servicewall.service'])
        return retval.decode().strip()
    except FileNotFoundError:
        raise AssertionError("'systemctl' not found in the path, is systemd"
                             " installed on this machine ?")
    except subprocess.CalledProcessError:
        # systemctl returns an 'error' on is-enabled if the answer is False.
        pass


def _enable_in_systemd():
    return _systemctl('enable')


def _disable_in_systemd():
    return _systemctl('disable')


class ServiceWall(statefulfirewall.StateFulFireWall):
    """ServiceWall - a FireWall in which you can add services on the fly.

    This class supersedes StateFulFireWall and adds services and realms.
    self.service_defs is a dict containing service definitions stored in
    self.service_defs_pickle, additioned with defs in self.service_defs_dir
    A service def file should be a json definition that is a valid ServiceDef.
    You can get a valid sample with eg

        sw show service http

    You can try yours with

        import json
        from servicewall.service_helpers import ServiceDef
        with open(your_file, "rb") as fd:
            service_dict = json.load(fd)
            print(ServiceDef(service_dict))
    """
    config_file = "/etc/servicewall/config.cfg"
    lib_dir = "/usr/lib/servicewall/"
    realm_defs_dict = "/etc/servicewall/realms.json"
    service_defs_pickle = "/usr/lib/servicewall/services.p"
    service_defs_dir = "/etc/servicewall/services/"
    dispatcher_toggler = "toggler"
    dispatchers = {
        "Network Manager": "/etc/NetworkManager/dispatcher.d/",
        "systemd-networkd": "/etc/networkd-dispatcher/carrier.d/",
    }

    def __init__(self):
        super().__init__()
        with open(self.realm_defs_dict, "r") as fd:
            self.realm_defs = json.load(fd)
        with open(self.service_defs_pickle, "rb") as fd:
            self.service_defs = pickle.load(fd)
        for service_file in os.listdir(self.service_defs_dir):
            if service_file.find('.') == 0:
                # It's a dotfile, keep going
                continue
            try:
                with open(self.service_defs_dir + service_file) as fd:
                    service_def = json.load(fd)
                    service_def["ports"] = PortDef(**service_def["ports"])
                    sdef = ServiceDef(**service_def)
                    self.service_defs[sdef.title] = sdef
            except TypeError as error:
                print("Warning : ServiceDef." + str(error),
                      "in " + self.service_defs_dir + service_file)
        try:
            self.realm_id = network_helpers.get_realm_id()
            self.online = True
        except KeyError:    # We don't have any network connection.
            self.realm_id = None
            self.online = False

        if self.online:
            self.subnetwork = network_helpers.get_subnetwork()
        else:
            self.subnetwork = False


    def start(self, should_check_hook=True, **args):
        """Will load a set of rules from self.realm_defs .
        """
        if should_check_hook:
            self._enable_hook()
        if self.realm_id not in self.realm_defs:
            # If we don't have a realm definition, load "ServiceWall:default"
            self.realm_defs[self.realm_id] = copy.deepcopy(
                self.realm_defs[self.identifier + ":default"])
        for service_name, scope in self.realm_defs[self.realm_id].items():
            self.insert_service_rule(service_name, scope=scope)
        # Commits the table if relevant, and brings other rules in :
        super().start(**args)

    def stop(self, should_check_hook=True):
        if should_check_hook:
            self._disable_hook()
        super().stop()

    def reload(self):
        self.stop(should_check_hook=False)
        self.start(should_check_hook=False)
        print("%s reloaded" % self.identifier)

    def enable(self):
        try:
            _enable_in_systemd()
        except AssertionError:
            with open(self.config_file, 'w+') as fd:
                fd.write('{ "state": "enabled" }')

    def disable(self):
        try:
            _disable_in_systemd()
        except AssertionError:
            with open(self.config_file, 'w+') as fd:
                fd.write('{ "state": "disabled" }')

    def is_enabled(self):
        try:
            status = _systemctl('is-enabled')
        except AssertionError:
            with open(self.config_file, 'r') as fd:
                config = json.load(fd)
                status = config["state"]
        if status != 'enabled':
            return False
        return True

    def allow_service(self, service_name, scope="local", realm=None):
        if service_name not in self.service_defs:
            raise KeyError("undefined service : %s." %
                           service_name)
        if realm is None:
            realm = self.realm_id
        # Create an entry for this realm's id if there weren't any :
        if realm not in self.realm_defs:
            self.realm_defs[realm] = copy.deepcopy(
                self.realm_defs[self.identifier + ":default"])
        if service_name not in self.realm_defs[realm]:
            self.realm_defs[realm][service_name] = scope
            self.save_rules()
            print("added %s to realm def %s" %
                  (service_name, realm))
            # We cannot hot-load it, there's an order on rules, best is to reload :
            if realm == self.realm_id:
                self.reload()

    def insert_service_rule(self, service_name, scope="local"):
        """Open ports for a service hosted on this machine.

        service_name should be one of self.service_defs' keys.
        if src is "local", use self.subnetwork instead.
        """
        if scope == "local":
            src = self.subnetwork
        elif scope == "docker":
            src = "172.16.0.0-172.31.255.255"
        else:
            src = ""

        for rule in self.input_chain.rules:
            if super()._get_rule_name(rule) == service_name:
                raise KeyError("rule already in input chain")

        service = self.service_defs[service_name]
        print("allowing service %s from %s" %
              (service_name, src or "anywhere"))
        for port in service.ports.tcp:
            rule = self.create_rule(service_name,
                                    "ACCEPT",
                                    src=src,
                                    dport=port,
                                    proto="tcp"
                                    )
            self.input_chain.insert_rule(rule)
        for port in service.ports.udp:
            rule = self.create_rule(service_name,
                                    "ACCEPT",
                                    src=src,
                                    dport=port,
                                    proto="udp"
                                    )
            self.input_chain.insert_rule(rule)

    def disallow_service(self, service_name, realm=None):
        if realm is None:
            realm = self.realm_id
        # Create an entry for this realm's essid if there weren't any :
        if realm not in self.realm_defs:
            self.realm_defs[realm] = copy.deepcopy(self.realm_defs[self.identifier + ":default"])
        # Do our own validity testing
        if service_name not in self.service_defs:
            raise KeyError('service "%s" not found.')
        if service_name in self.realm_defs[realm]:
            del self.realm_defs[realm][service_name]
            self.save_rules()
            print("removed service %s from realm %s" %
                  (service_name, realm))
        else:
            raise KeyError('service "%s" was not allowed in realm %s anyway.'
                           % (service_name, realm))
        if realm == self.realm_id:
            self.remove_service_rule(service_name)

    def remove_service_rule(self, service_name):
        """Closes ports for service service_name if they were opened.
        """
        for rule in self.input_chain.rules:
            # Call the FireWall's  private _get_rule_name function
            if super()._get_rule_name(rule) == service_name:
                self.del_rule(service_name, self.input_chain)
                #self.input_chain.delete_rule(rule)

    def save_rules(self):
        with open(self.realm_defs_dict, "w") as fd:
            json.dump(self.realm_defs, fd, indent=2)
        print("saved realm defs to config")

    def list_services_in(self):
        """Lists services for which we have allowed ports.
        """
        self.list_rules(self.input_chain)

    def list_services_by_port(self, port):
        services_list = []
        for service_name, s_tuple in self.service_defs.items():
            for service_port_range in (*s_tuple.ports.tcp, *s_tuple.ports.udp):
                # service_port_range is a string containing either a number or
                # a range, as in "80:88", "120"
                if service_port_range.isalnum():
                    if port == service_port_range:
                        services_list.append(service_name)
                else:   # it's a range
                    start, end = service_port_range.split(":")
                    if port in range(int(start), int(end)+1):
                        services_list.append(service_name)
        return services_list

    def _enable_hook(self):
        """Create a link in the network dispatcher pointing to the event triggerer.
        """
        if not os.path.exists(self.lib_dir):
            # Should be installed by setup.py .
            raise SystemExit("Could not find %s in %s. Check your installation"
                             % (self.dispatcher_toggler, self.lib_dir))
        # We will only mark as enabled if we can link to a network dispatcher :
        linked = False
        for dispatcher, dst_dir in self.dispatchers.items():
            if not os.path.exists(dst_dir):
                # Keep going with the next dispatcher.
                continue
            if os.path.exists(dst_dir + self.dispatcher_toggler):
                print("%s dispatcher was already linked" % dispatcher)
                linked = True
            else:
                print("%s dispatcher link created" % dispatcher)
                # symlink pointing to src in dst_dir
                os.symlink(self.lib_dir + self.dispatcher_toggler,
                           dst_dir + self.dispatcher_toggler)
                linked = True
        if not linked:
            raise SystemExit("Could not link to any network event dispatcher. "
                             "You apparently aren't running neither Network "
                             "Manager nor systemd-networkd with "
                             "networkd-dispatcher. You'll need one of those to"
                             " run this as it relies on them to fire the"
                             "network change events.")

    def _disable_hook(self):
        """Destroy the link in the network dispatcher pointing to the event triggerer.
        """
        for dispatcher, target in self.dispatchers.items():
            # DEBUG This test would fail on a broken link :
            if os.path.exists(target + self.dispatcher_toggler):
                os.remove(target + self.dispatcher_toggler)
                print("%s dispatcher link destroyed" % dispatcher)
            else:
                # Report missing link only if dir is present
                if os.path.exists(target):
                    print("there was no %s dispatcher link" % dispatcher)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Helper functions for ServiceWall cmdline agent.
"""

__all__ = [ "no_arg_provided", "enable", "disable", "start", "stop", "reload",
            "show_logs", "show_realm", "show_realms", "show_services",
            "show_service", "status", "allow_service", "disallow_service" ]


import json
from datetime import datetime
#import select
#import putch
import socket
import servicewall


DEBUG = False
firewall = servicewall.ServiceWall()


def prettyprint(obj):
    print(json.dumps(obj, indent=2))


def print_dict(dictionary, depth=1):
    prefix = "  " * depth
    for key, value in dictionary.items():
        if isinstance(value, dict):
            print("%s=> %s :" % (prefix, key))
            print_dict(value, depth+1)
        else:
            print("%s%s : %s" % (prefix, key, value))


def print_service(port):
    services_list = firewall.list_services_by_port(port)
    if len(services_list) > 1:
        return '%s*' % services_list[0]
    elif len(services_list) == 1:
        return '%s' % services_list[0]
    return None


def no_arg_provided(args):
    parser.print_help()
    #raise SystemExit("\n  argument needed !")


def enable(args):
    firewall.enable()


def disable(args):
    firewall.disable()


def start(args):
    firewall.start()


def stop(args):
    firewall.stop()


def reload(args):
    firewall.reload()


def status(args):
    if firewall.is_enabled():
        print("ServiceWall is enabled")
    else:
        print("ServiceWall is disabled")
    if firewall.up:
        if firewall.realm_id:
            realm_name = firewall.realm_id
        else:
            realm_name = "no network"
        print("and started - using profile for realm %s" % realm_name)
    else:
        print("and stopped")


def allow_service(args):
    if args.globally:
        scope = "global"
    elif args.docker:
        scope = "docker"
    else:
        scope = "local"
    if args.in_default_profile:
        realm = "ServiceWall:default"
    else:
        realm = None
    # Make it local by default :
    firewall.allow_service(args.service_name, scope=scope, realm=realm)


def disallow_service(args):
    service_name = args.service_name
    if args.in_default_profile:
        realm = "ServiceWall:default"
    else:
        realm = None
    firewall.disallow_service(service_name, realm=realm)


def show_table(args):
    print('You are using realm profile : %s' %
          (firewall.realm_id or "ServiceWall:default"))
    print('=> Input rules <=           : policy %6s' %
          firewall.input_chain.get_policy().name)
    for rule in firewall.list_rules(firewall.input_chain):
        print(rule)
    print('=> Forward rules <=         : policy %6s' %
          firewall.forward_chain.get_policy().name)
    for rule in firewall.list_rules(firewall.forward_chain):
        print(rule)
    print("=> Output rules <=          : policy %6s" %
          firewall.output_chain.get_policy().name)


def show_realm(args):
    if firewall.realm_id:
        print('This machine is connected to ESSID "%s".' % firewall.realm_id)
    else:
        print('This machine is not connected.')
    #prettyprint(firewall.realm_defs[firewall.realm_id])


def show_realms(args):
    #print_dict(firewall.realm_defs)
    prettyprint(firewall.realm_defs)


def show_service(args):
    s = firewall.service_defs[args.service_name]._asdict()
    s["ports"] = s["ports"]._asdict()
    prettyprint(s)


def show_services(args):
    for service in firewall.service_defs:
        print("%s - %s"
              % (service, firewall.service_defs[service].description))


def show_port(args):
    port = args.port_name
    services_list = firewall.list_services_by_port(port)
    if services_list:
        if len(services_list) == 1:
            print("service using port %s : %s" % (port, services_list[0]))
        else:
            print('services using port %s :' % port)
            #[ print("  - %s" % i) for i in services_list ]
            prettyprint(services_list)
    else:
        print("port %s unknown." % port)


def show_logs(args):
    if "period" in args:
        yielder = firewall.yield_logs(period=args.period)
    else:
        yielder = firewall.yield_logs()
    if "number" in args:
        limit = int(args.number)
        i = 0
    else:
        limit = None

    now = datetime.today()

    for log in yielder:
        if limit:
            i += 1
            if i > limit:
                break

        if "with_hostnames" in args:
            # If we have a hostname, output it - shortened if longer than :
            try:
                hostname = socket.gethostbyaddr(log['SRC'])[0]
                if len(hostname) > 20:
                    log['SRC'] = '%s ..%s' % (log['SRC'], hostname[-17:])
                else:
                    log['SRC'] = '%s %s' % (log['SRC'], hostname)
            except socket.herror:
                pass

        # If we have ports, try to name the associated service :
        if log['DPT'] or log['SPT']:
            servicename = print_service(log['DPT'])
            if servicename:
                service = '> %s (%s/%s)' % (
                    servicename,
                    log['PROTO'],
                    log['DPT']
                )
            else:
                servicename = print_service(log['SPT'])
                if servicename:
                    service = '>>%s (%s/%s)' % (
                        servicename,
                        log['PROTO'],
                        log['SPT']
                    )
                else:
                    # We have no port, display protocol :
                    service = '%5s>%5s %-5s' % (
                        log['SPT'],
                        log['DPT'],
                        log['PROTO']
                    )
        else:
            # We have no port, display protocol :
            service = 'layer3: %s' % log['PROTO']

        age = now - log['LOG_DATE']
        if age.days:
            age = str(age).split(',')[0]
        else:
            age = str(age).split('.')[0]

        print('%-36s %-16s %17s %-8s' % (
            log['SRC'],
            log['DST'],
            service,
            age
        ))




"""Useful system calls to ask different things to the wireless controler.

Most of them use fnctl.ioctl to pack and unpack a request.

some insights on packing an address :
https://pymotw.com/2/socket/addressing.html

For those taking an interface name iface as argument, see struct ifreq in
    /usr/include/net/if.h
we'll mimic ifreq with struct.pack("256s", bytes(ifname[:15], "utf-8"))
details on constructing a C struct in python :
https://docs.python.org/2/library/ctypes.html#arrays
"""


import sys
import socket
import struct
import array
import fcntl


def get_all_interfaces():
    """Return a dict of interface names and their IPs.

    retrieved from http://code.activestate.com/recipes/439093/#c8
    """
    is_64bits = sys.maxsize > 2**32
    struct_size = 40 if is_64bits else 32
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    max_possible = 8 # initial value
    while True:
        _bytes = max_possible * struct_size
        names = array.array('B')
        for i in range(0, _bytes):
            names.append(0)
        outbytes = struct.unpack('iL', fcntl.ioctl(
            datagram_socket.fileno(),
            0x8912,  # SIOCGIFCONF
            struct.pack('iL', _bytes, names.buffer_info()[0])
        ))[0]
        if outbytes == _bytes:
            max_possible *= 2
        else:
            break
    namestr = names.tobytes()
    ifaces = []
    for i in range(0, outbytes, struct_size):
        iface_name = bytes.decode(namestr[i:i+16]).split('\0', 1)[0]
        iface_addr = socket.inet_ntoa(namestr[i+20:i+24])
        ifaces.append((iface_name, iface_addr))
    # We don't care for the loopback interface :
    for index, iface in enumerate(ifaces):
        if iface[0] == "lo":
            del ifaces[index]
    return ifaces


def get_active_interface():
    """Return the name of the interface serving as gateway.
    """
    # AF_INET is a flag for ipv4 addressing.
    #return netifaces.gateways()[netifaces.AF_INET][0][1]
    return get_all_interfaces()[0][0]


def get_essid():
    """Return the ESSID for an interface, or None if we aren't connected.
    """
    try:
        interface = get_active_interface()
    except IndexError:
        #print("No active interface")
        return None
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # The payload for this call is formatted to 32 double hexadecimal ints.
    essid = array.array("b", b"\x00" * 32)
    essid_pointer, essid_length = essid.buffer_info()
    request = array.array("b")
    request.frombytes(
        # The interface name is formatted to 16 double hexadecimal ints.
        interface.ljust(16, "\x00").encode() +
        struct.pack("PHH", essid_pointer, essid_length, 0)
        )
    try:
        fcntl.ioctl(
            datagram_socket.fileno(),
            0x8b1b,   # SIOCGIWESSID, get essid
            request
            )
    except OSError as error:
        if error.errno == 95:       # Operation not supported
            #print("not connected")
            return None
    name = essid.tobytes().strip(b"\x00").decode()
    if not name:
        return None
    return name


def get_realm_id():
    """Return a unique identifier for the Internet Service Provider
    """
    isp_id = get_essid()
    if not isp_id:
        # Then the network probably isn't wifi. Check ethernet :
        import arpreq
        #isp_id = arpreq.arpreq(get_ip_address(get_active_interface()))
        isp_id = arpreq.arpreq(get_gateway_address())
    return isp_id


def get_ip_address(ifname):
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        datagram_socket.fileno(),
        0x8915,  # SIOCGIFADDR
        struct.pack("256s", bytes(ifname[:15], "utf-8"))
    )[20:24])


def get_outside_ip_address(ifname):
    """DOESN'T WORK and I don't know what is this IP it's outputting
    """
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        datagram_socket.fileno(),
        0x8927,  # SIOCGIFHWADDR
        struct.pack("256s", bytes(ifname[:15], "utf-8"))
    )[20:24])


def get_netmask(ifname):
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        datagram_socket.fileno(),
        0x891b,  # SIOCGIFNETMASK
        struct.pack("256s", bytes(ifname[:15], "utf-8"))
    )[20:24])


def get_subnetwork():
    """return the IP address of the subnetwork slash the netmask, as in

            "192.168.1.0/255.255.255.0"
    """
    interface = get_active_interface()
    address = get_ip_address(interface).split(".")
    netmask = get_netmask(interface).split(".")
    subnetwork = [ str(int(int(address_block) * int(netmask_block) / 255))
                   for address_block, netmask_block in zip(address, netmask) ]
    return "/".join((".".join(subnetwork), ".".join(netmask)))


def get_essid_mac_address(ifname):
    """Return the ssid's MAC address as declared by the network controller.
    DEBUG Doesn't work with ethernet network providers though.
    """
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    payload = fcntl.ioctl(
        datagram_socket.fileno(),
        0x8b15,  # SIOCGIWAP , mac address
        struct.pack("265s", bytes(ifname[:15], "utf-8"))
    )
    # From internet talk : SSID is encoded after a \x01\x00 keycode.
    index = payload.index(b"\x01\x00") + 2
    # From convention : MAC addressing is 6 hexadecimal double ints.
    address = payload[index:index+6].hex()
    return ":".join([ address[i:i+2] for i in range(0, 12, 2) ])


def get_mac_address(ip):
    """NOT WORKING - need to find how to mimic arpreq struct - clues in
    https://stackoverflow.com/questions/159137/getting-mac-address
    example of what's expected, written in C :
    https://stackoverflow.com/questions/11929556/how-to-get-the-mac-address-of-client-using-ioctl-function
    """
    datagram_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    interface = get_active_interface()
    # SIOCGARP is documented in man 7 arp
    # and struct arpreq in /usr/include/net/if_arp.h .
    # It refers to struct sockaddr, defined in /usr/inlude/sys/socket.h .
    request = array.array("b")
    request.frombytes(
        interface.ljust(16, "\x00").encode()
    )
    return socket.inet_ntoa(fcntl.ioctl(
        datagram_socket.fileno(),
        0x8951,  # SIOCGARP , mac address, according to man ioctl_list
        #0x8954, # SIOCGARP , mac address
        request
    )[20:24])


def get_essid_alt(ifname):
    from subprocess import Popen, PIPE
    process = Popen(["iw", ifname, "info"], stdout=PIPE)
    stdout, _ = process.communicate()
    try:
        endstring = stdout[stdout.index(b"ssid ") + 5:]
    except ValueError:
        raise SystemExit("Interface %s not found" % ifname)
    ssid = endstring.split(b"\n")[0]
    return ssid.decode()


def get_gateway_address():
    """Read the default gateway directly from /proc.
    """
    with open("/proc/net/route") as fh:
        address = None
        for line in fh:
            fields = line.strip().split()
            if fields[1] != '00000000' or not int(fields[3], 16) & 2:
                continue
            address = socket.inet_ntoa(struct.pack("<L", int(fields[2], 16)))
        if not address:
            raise KeyError("No gateway found")
    return address


def get_gateway_hostname():
    return socket.getfqdn(get_gateway_address())




"""Firewall class

implements a firewall using python-iptables
"""

from iptc import Rule, Rule6, Chain, Table, Table6
from iptc.ip4tc import IPTCError


class FireWall():
    """A simple firewall class

    In this firewall, there is (only) one _table called FILTER. It contains
    2 chains called input and output. When anything comes from outside, it
    goes through input. When anything goes outside, it goes through output.

    You can add (and delete) rules based on any valid iptables match.

    TODO :
    - ruleset Object.
    """

    identifier = "ServiceWall"

    def __init__(self):
        try:
            self._table = Table(Table.FILTER)
            self._table6 = Table6(Table6.FILTER)
        except IPTCError:
            # user is not root - return silently ; all writing ops will fail
            return

        # _table.autocommit is True by default.
        #self._table.autocommit = False
        self.input_chain = Chain(self._table, "INPUT")
        self.forward_chain = Chain(self._table, "FORWARD")
        self.output_chain = Chain(self._table, "OUTPUT")
        self.input_chain6 = Chain(self._table6, "INPUT")
        self.forward_chain6 = Chain(self._table6, "FORWARD")
        self.output_chain6 = Chain(self._table6, "OUTPUT")
        if self.status():
            self.up = True
        else:
            self.up = False

    def start(self):
        """Start a basic FireWall.

        Drops all incoming, accepts only localhost.
        """
        # Drop all incoming, allow outgoing requests.
        print("setting input policy to DROP")
        self.input_chain.set_policy("DROP")
        print("setting forward policy to DROP")
        self.forward_chain.set_policy("DROP")
        print("setting output policy to ACCEPT")
        self.output_chain.set_policy("ACCEPT")
        print("disabling ipv6 stack")
        self.input_chain6.set_policy("DROP")
        self.forward_chain6.set_policy("DROP")
        self.output_chain6.set_policy("DROP")
        # Accept all from _and_ to localhost.
        accept_localhost_rule = self.create_rule(
            "localhost",
            "ACCEPT",
            #siface="lo"
            src="127.0.0.1",
            dst="127.0.0.1"
            )
        self.input_chain.insert_rule(accept_localhost_rule)
        if not self._table.autocommit:
            self._table.commit()
        self.up = True

    def stop(self):
        """Remove rules.

        This function removes all rules created by this firewall's instances ;
        it recognizes them by the comment that's inside of them.
        """
        print("flushing rules")
        found_rules = 0
        for rule in self.input_chain.rules:
            if self._get_rule_id(rule) == self.identifier:
                self.del_rule(self._get_rule_name(rule), self.input_chain)
                found_rules += 1
        for rule in self.forward_chain.rules:
            if self._get_rule_id(rule) == self.identifier:
                self.del_rule(self._get_rule_name(rule), self.forward_chain)
                found_rules += 1
        if not found_rules:
            print("no rule found for id %s." % self.identifier)
        print("setting input policy to ACCEPT")
        self.input_chain.set_policy("ACCEPT")
        print("setting forward policy to ACCEPT")
        self.forward_chain.set_policy("ACCEPT")
        if not self._table.autocommit:
            self._table.commit()
        self.up = False

    def status(self):
        """print the status of the FireWall. Returns either True or False."""
        # Returns True if a single rule has our identifier tag
        is_up = False
        for rule in self.input_chain.rules:
            if self._get_rule_id(rule) == self.identifier:
                is_up = True
        return is_up

    def create_rule(self, name, target, dst="", dport="", src="", sport="", proto="", siface="", diface=""):
        """create and return a rule (or two if no proto given).

        name can be any string
        target is a string saying what to do, like
                "ACCEPT", "DENY", "DROP", "LOG"
        proto (if not defined, is assumed both udp and tcp)
        src is a realm in the form XXX.XXX.XXX.XXX/YY
        sport is source port. Currently ignored
        dst is a realm in the same form as src
        dport is destination port
        siface is the source interface
        diface is the destination interface
        """
        #print("adding rule %s" % name)
        if (dport or sport) and not proto:
            return [ self.create_rule(name,
                                      target,
                                      dst,
                                      dport,
                                      src,
                                      sport,
                                      proto,
                                      siface,
                                      diface) for proto in ("tcp", "udp") ]
        rule = Rule()
        rule.create_target(target)
        if dst:
            rule.dst = dst
        if dport:
            proto_match = rule.create_match(proto)
            proto_match.dport = str(dport)
            rule.protocol = proto
        if diface:
            rule.in_interface = diface
        if src:
            if src.count("-"):
                iprange_match = rule.create_match("iprange")
                iprange_match.src_range = src
                rule.add_match(iprange_match)
            else:
                rule.src = src
        if sport:
            proto_match = rule.create_match(proto)
            proto_match.sport = str(sport)
            rule.protocol = proto
        if siface:
            rule.out_interface = siface
        # Add a signature as a comment.
        comment_match = rule.create_match("comment")
        comment_match.comment = self.identifier + ":" + name
        rule.final_check()
        return rule

    def del_rule(self, name, chain):
        """Remove rule.

        This function removes one rule named "name" in chain "chain". It
        expects the rule to contain a comment in the form
            "identifier:name"
        where identifier is the name of the program that has set the rule, and
        title is the name of the rule. Note that there may be several rules
        under the same name. This should delete the first it finds.
        """
        for rule in chain.rules:
            if self._get_rule_name(rule) == name:
                chain.delete_rule(rule)
                #print("deleted rule %s" % name)
                break

    def list_rules(self, chain):
        for rule in chain.rules:
            # Find out dport - it's set by self.add_service_in as a match.
            dport = "any port"
            proto = "tcp/udp"
            for match in rule.matches:
                try:
                    dport = "port " + match.parameters["dport"]
                    proto = match.name
                except (KeyError, IndexError):
                    pass
            # Then find out source.
            if rule.src == "0.0.0.0/0.0.0.0":
                src = "any source"
            else:
                src = rule.src
            yield "%6s %-20s : %7s %-8s from %-23s" % (
                rule.target.name, self._get_rule_name(rule), proto, dport, src
            )

    def _get_rule_name(self, rule):
        for match in rule.matches:
            if match.comment:
                try:
                    _, rule_name = match.comment.split(":")
                except IndexError:
                    # No semicolon in comment
                    continue
                return rule_name

    def _get_rule_id(self, rule):
        for match in rule.matches:
            if match.comment:
                try:
                    rule_identifier, _ = match.comment.split(":")
                except IndexError:
                    # No semicolon in comment
                    continue
                return rule_identifier




#!/usr/bin/env python
"""Install script for servicewall
"""


import sys
import stat
from os import path, chmod, popen
from distutils import log
import setuptools
from setuptools.command.install import install
import _version


class CustomInstall(install):
    """Only let user write and group read files in /etc/servicewall"""
    def run(self):
        mode = stat.S_IWRITE + stat.S_IREAD + stat.S_IRGRP
        dirmode = stat.S_IWRITE + stat.S_IREAD + stat.S_IXUSR + stat.S_IRGRP + stat.S_IXGRP
        install.run(self)
        for filepath in self.get_outputs():
            if "/etc/servicewall" in filepath:
                log.info("setting %s to mode %s" % (filepath, oct(mode)[2:]))
                dirpath = path.split(filepath)[0]
                chmod(filepath, mode)
                chmod(dirpath, dirmode)


with open("README.md", "r") as fd:
    long_description = fd.read()

NAME = "servicewall"
here = path.abspath(path.dirname(__file__))

for package in setuptools.find_packages():
    print("setuptools : including package %s" % package, file=sys.stderr)

setuptools.setup(
    name=NAME,
    version=_version.version,
    author="la Fleur",
    author_email="lafleur@boum.org",
    description="The firewall that remembers the different network you connect to.",
    keywords="dynamic adaptable iptables firewall",
    license="GNUv3",
    python_requires=">=3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.relatif.moi/lafleur/servicewall",
    packages=setuptools.find_packages(),
    #packages=setuptools.find_packages(where="src"),
    classifiers=[
        'Development Status :: 4 - Beta',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        #"Operating System :: OS Independent",
        "Operating System :: POSIX",   # We need iptables.
    ],
    install_requires=[
        "argparse",
        "arpreq",
        "python-iptables",
        "python-systemd",
    ],
    extras_require={
        "argument completion as root": "argcomplete",
    },
    scripts=[
        "servicewall/sw",
    ],
    data_files=[
        ("/etc/systemd/system", [
            "etc/systemd/servicewall-ulogd.service",
            "etc/systemd/servicewall-logs.service",
            "etc/systemd/servicewall-logs.socket",
            "etc/systemd/servicewall.service",
        ]),
        ("/etc/servicewall/", [
            "etc/servicewall/realms.json",
            "etc/servicewall/ulogd.conf",
        ]),
        ("/etc/servicewall/services", [
            "etc/servicewall/services/.keepme",
        ]),
        ("/etc/xdg/autostart/", [
            "lib/servicewall-systray.desktop",
        ]),
        ("lib/servicewall/", [
            "lib/systray.py",
            "lib/services.p",
            "lib/toggler",
            "lib/icon.png",
            "lib/icon2.png",
        ]),
    ],
    cmdclass={"install": CustomInstall},
)




"""  StateFulFireWall

Uses simple guidelines given in

and implements them in a FireWall class, using reasonable defaults.
"""

from datetime import datetime
import socket
from systemd import journal
from iptc import Rule
from servicewall import firewall


class StateFulFireWall(firewall.FireWall):
    """Implement some useful stateful rules :

    - accept related/established packets
    - log anything that is dropped
    - drop invalid packets
    """
    protobynumber = { num: name[8:].lower()
                      for name, num in vars(socket).items()
                      if name.startswith("IPPROTO")
                      }

    def start(self, **args):
        print("adding rule icmp")
        rule = Rule()
        rule.create_target("ACCEPT")
        rule.protocol = "icmp"
        icmp_match = rule.create_match("icmp")
        icmp_match.set_parameter("icmp-type", "8")
        comment_match = rule.create_match("comment")
        comment_match.comment = self.identifier + ":icmp"
        self.input_chain.append_rule(rule)

        # Top rule should be to allow related, established connections.
        related_rule = self.create_conntrack_rule("ACCEPT",
                                                  "ctstate",
                                                  "RELATED,ESTABLISHED")
        self.input_chain.insert_rule(related_rule)

        # Drop igmp packets before logging.
        drop_igmp_rule = Rule()
        drop_igmp_rule.create_target("DROP")
        drop_igmp_rule.protocol = "igmp"
        self.input_chain.append_rule(drop_igmp_rule)
        # Log all that is refused in INPUT chain.
        log_rule = self.create_log_rule("not in allowed services", group="1")
        self.input_chain.append_rule(log_rule)

        # Drop invalid packets - as diagnosed by the conntrack processor.
        # Note that this rule is useless because packets would be dropped anyway.
        invalid_rule = self.create_conntrack_rule("DROP", "ctstate", "INVALID")
        self.input_chain.append_rule(invalid_rule)

        super().start(**args)   # commits the table if relevant

    def create_conntrack_rule(self, target, param_key, param_value):
        """creates a connection tracking rule, for example :

        StateFulFireWall.create_conntrack_rule("ACCEPT", "ctstate", "RELATED")
        """
        print("adding rule %s" % param_value)
        conntrack_rule = Rule()
        conntrack_rule.create_target(target)
        conntrack_match = conntrack_rule.create_match("conntrack")
        conntrack_match.set_parameter(param_key, param_value)
        comment_match = conntrack_rule.create_match("comment")
        comment_match.comment = self.identifier + ":" + param_value
        return conntrack_rule


    def create_log_rule(self, comment="", group="", dst="", dport="", src="",
                        sport="", proto="", iface=""):
        """Adds a log rule

        All arguments should be strings !
        The comment size is limited to 64 characters as a whole.
        The nflog_group is the one you will have to select in ulogd.
        """
        log_rule = self.create_rule("log", "NFLOG", dst, dport, src, sport,
                                    proto, iface)
        limit_match = log_rule.create_match("limit")
        limit_match.limit = "1/s"
        limit_match.limit_burst = "1"
        if comment:
            log_rule.target.set_parameter("nflog-prefix", comment)
        if group:
            log_rule.target.set_parameter("nflog-group", str(group))
        return log_rule


    def yield_logs(self, limit=None, period=None):
        """get logs we implemented in iptables from journald

        limit [int] is the max number of logs to yield
        period [int] is the age in seconds of the oldest log to yield
        """
        reader = journal.Reader()
        #reader.log_level(journal.LOG_WARNING)
        # That would be for LOG match :
        #reader.add_match(SYSLOG_IDENTIFIER="kernel")
        # That is for ulog systemd service associated to ulog.socket :
        reader.add_match(_SYSTEMD_UNIT="servicewall-logs.service")
        now = datetime.now()
        hostname = socket.gethostname()
        #p = select.poll()
        #p.register(reader, reader.get_events())
        #p.poll()
        reader.seek_tail()
        if limit:
            limit = int(limit)
            i = 1
        if period:
            period = int(period)
        while True:
            log = reader.get_previous()
            if "MESSAGE" not in log:
                continue
            # Quit if log older than period :
            if period:
                age = now - log["__REALTIME_TIMESTAMP"]
                if int(age.total_seconds()) > period:
                    break
            message_dict = {}
            # Get rid of the trailing date and hostname :
            try:
                message = log["MESSAGE"].split(hostname)[-1].strip()
            except IndexError:
                print("!! error :")
                print(log["MESSAGE"])

            # If message begins with "[NEW] "or "[DESTROY]", then it is a
            # conntrack info log - skip.
            if message.startswith("[NEW]") or message.startswith("[DESTROY]"):
                continue

            # Only count the logs that make it here :
            if limit:
                if i < limit:
                    i += 1
                else:
                    break

            message_dict["LOG_DATE"] = log["__REALTIME_TIMESTAMP"]
            for item in message.split():
                if item.count("="):
                    key, value = item.split("=")
                    message_dict[key] = value
                else:
                    message_dict[item] = ""

            if 'DPT' not in message_dict:
                message_dict['DPT'] = ''
            if 'SPT' not in message_dict:
                message_dict['SPT'] = ''
            if message_dict['PROTO'].isnumeric():
                message_dict['PROTO'] = self.protobynumber[
                        int(message_dict['PROTO'])]
            else:
                message_dict['PROTO'] = message_dict['PROTO'].lower()

            yield message_dict

    def filter_logs_by(self, criteria, limit=None, period=None):
        """output logs sorted by a log variable, like "DPT" or "SRC"

        note that limit applies to the criteria (like sort nth first SRC hosts)
        """
        yielder = self.yield_logs(period=period)
        logs = {}
        if limit:
            limit = int(limit)
            i = 0
        if period:
            period = int(period)
        for log in yielder:
            if log[criteria] not in logs:
                logs[log[criteria]] = [log, ]
                if limit:
                    if i < limit:
                        i += 1
                    else:
                        # Stop at step i+1 and pop last record :
                        logs.popitem()
                        break
            else:
                logs[log[criteria]].append(log)
        return logs

